//=======================================================================
// Basic C++: quelques extensions utiles de la STL
//-----------------------------------------------------------------------
// Jean-Paul Rigault --- Copyright 2003
// $Id: stl_extensions.h,v 1.1 2005/05/01 16:58:16 jpr Exp $
//=======================================================================

#ifndef _STL_EXTENSIONS_H_
#define _STL_EXTENSIONS_H_

#include <iostream>
#include <iterator>
#include <algorithm>
#include <string>
using namespace std;

namespace JPR_STL_EXTENSIONS 
{

//-----------------------------------------------------------------------------
// Extensions  struct pair<>
//-----------------------------------------------------------------------------

// Premire et seconde projection des paires
template <typename A, typename B>
A proj1(const pair<A, B>& p)
{
    return p.first;
}

template <typename A, typename B>
B proj2(const pair<A, B>& p)
{
    return p.second;
}

// Il est commode de doter les paires d'un oprateur d'affichage
template <typename A, typename B>
ostream& operator<<(ostream& os, const pair<A, B>& p)
{
    return os << '(' << p.first << ", " << p.second << ')';
}

//-----------------------------------------------------------------------------
// Affichage des collections
//-----------------------------------------------------------------------------

// Les collections (containers) de la STL ne disposent pas, par dfaut, de
// l'oprateur << des ostreams. On peut essayer de leur en donner un (template
// bien sr), mais il y a quelque problmes dlicats.
//
// Si on essaie quelque chose comme
//
//	template <typename Container> 
// 	ostream& operator<<(ostream& os, const Container& cont);  
//
// cela marche... mais trop bien ! En effet cette dfinition est applicable
// pourContainer valant n'importe quel type, et pas seulement une collection de
// la STL !
//
// On peut essayer alors
//
//	template <typename T, template <typename> class Container> 
//	ostream& operator<<(ostream& os, const Container<T>& cont);
//
// qui passe *toute une classe template en paramtre*. Cela marche
// effectivement pour tous les containers de la STL et donc, malheureusement,
// aussi pour Container = basic_string ! Cela dentre donc en conflit avec
// l'oprateur << prdfini des strings.
// 
// Nous avons donc recours  une fonction ordinaire, display(), qui marche pour tous les
// containers (et sans doute quelques autres classes !). Comme nous utilisons for_each() pour implmenter display(), nous avons besoin d'une fonction auxiliaire, utile aussi.

// Impression d'un objet quelconque, avec sparateur paramtrable
template <typename T>
class print_obj
{
private:
    ostream& _os;
    string _sep;

public:
    print_obj(ostream& os = cout, const string& sep = " ")
	: _os(os), _sep(sep)
    {}

    // Affichage  partir d'une instance ou rfrence
    void operator()(const T& t)
    {
	_os << t << _sep;
    }

    // Affichage  partir d'un pointeur
    void operator()(T* pt)
    {
	_os << *pt << _sep;
    }
};

// Affichage d'un container quelconque
template <typename Container>
void  display(ostream& os, const Container& cont, const string& sep = " ")
{
    typedef typename Container::value_type T;
    
    for_each(cont.begin(), cont.end(), print_obj<T>(os, sep));
}

//-----------------------------------------------------------------------------
// Quelques nouveaux objets-fonctions
//-----------------------------------------------------------------------------

// Une fonction gnrale et utile pour dtruire toutes sortes d'objets
// (pourvu qu'ils aient t allous par new, bien sr !)
// (cette fonction est peut-tre un peu avance en deuxime anne !!!)
struct delete_obj 
{ 
    // Voici une fonction-membre template, cad une fonction-membre surcharge
    // et dont la surcharge est paramtre : aucun mystre si vous y
    // rflchissez...
    template <typename Obj> 
    void operator()(Obj*& pobj) const 
    { 
	delete pobj; 
	pobj = 0; // on remet le pointeur  0 ce qui est pratiquement
                  // ncessaire dans la plupart des cas ; c'est pourquoi le
                  // pointeur est pass par rfrence
    } 
};

//-----------------------------------------------------------------------------
// Quelques nouveaux algorithmes
//-----------------------------------------------------------------------------

// Algorithme copy_if()
//
//	copy_if(first, last, target, pred)
//
// copie dans la collection commenant  target, tous les lments de la
// collection [first, last[ qui satisfont le prdicat pred. Retourne le nombre
// d'lments copis.
template <typename InputIterator, typename OutputIterator, typename Predicate>
int copy_if(InputIterator first, InputIterator last,
            OutputIterator target, Predicate pred) 
{
    InputIterator it;
    int n;

    for (n = 0, it = first; it != last; it++)
    {
	if (pred(*it)) 
        {
            *target++ = *it;
            ++n;
        }
    }
    return n;
}

// Algorithme for_each_if()
//
//	for_each_if(first, last, uf, pred)
// 
// applique la fonction uf() sur tous les lments de la collection [first, last[
// qui satisfont le prdicat pred(). Retourne une copie de uf().
template <typename InputIterator, typename Predicate, typename UnaryFunction>
UnaryFunction for_each_if(InputIterator first, InputIterator last, 
			  UnaryFunction uf, Predicate pred)
{
    InputIterator it;

    for (it = first; it != last; it++)
    {
	if (pred(*it)) uf(*it);
    }
    return uf;
}


//-----------------------------------------------------------------------------
} // find du namespace JPR_STL_EXTENSIONS 

#endif
