//=======================================================================
// Basic C++: classe Date
//      Implementation of class Date
//-----------------------------------------------------------------------
// Jean-Paul Rigault --- Copyright 2003
// $Id: Date.cpp,v 1.4 2004/11/07 10:52:52 jpr Exp $
//=======================================================================

#include <time.h>

#include "Date.h"

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------

// POSIX localtime does not handle date earlier than 1900!
const int MIN_YEAR = 1900;
// Number of months
const int NMONTHS = 12;
// Maximum number of days in a month
const int NDAYS = 31;

// Name of months
static const char *Month[] = {
    "", "January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"
};

// Number of days in months: Month[0] is for non leap year, 
// Month{1] for leap year
static const int Days[][NMONTHS + 1] = {
    {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
    {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
};

//-----------------------------------------------------------------------
// Static functions
//-----------------------------------------------------------------------

Date Date::today()
{
    return Date();
}

bool Date::is_leap(int y)
{
    if (y % 100 == 0) return y % 400 == 0;
    else return y % 4 == 0;
}


//-----------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------

// Default constructor: initialize today (POSIX way!)
Date::Date()
{
   time_t t = time(0);
   struct tm *ptm = localtime(&t);
   _year = ptm->tm_year + MIN_YEAR;
   _month = ptm->tm_mon + 1;
   _day = ptm->tm_mday;
}

// Normal constructor
Date::Date(int y, int m, int d)
{
    if (y < MIN_YEAR) throw Bad_Year();
    if (m <= 0 || m > NMONTHS) throw Bad_Month();
    if (d <= 0 || d > Days[is_leap(y)][m]) throw Bad_Day();

    _year = y;
    _month = m;
    _day = d;
}

//-----------------------------------------------------------------------
// Arithmetic operations
//-----------------------------------------------------------------------

// Add n days
Date Date::operator+(int n) const
{
    int y;
    int m;
    int d;
    int i;

    for (i = 0, d = _day, m = _month, y = _year; i < n; ++i, ++d)
    {
        if (d >= Days[is_leap(y)][m]) // we change month
        {
            d = 0; // will be incremented before next iteration
            ++m;
            if (m > NMONTHS) // we change year also
            {
                ++y;
                m = 1;
            }
        }
    }
    return Date(y, m, d);
}

// Backtrack n days
Date Date::operator-(int n) const
{
    int y;
    int m;
    int d;
    int i;

    for (i = n, d = _day, m = _month, y = _year; i > 0; --i, --d)
    {
        if (d == 1) // we change month
        {
            --m;
            if (m <= 0) // we change year also
            {
                if (--y < MIN_YEAR) throw Bad_Year();
                m = NMONTHS;
            }
            d = Days[is_leap(y)][m] + 1;
                        // will be decremented before next iteration
        }
    }
    return Date(y, m, d);
}

// Number of days between two dates
int operator-(Date dat1, Date dat2)
{
    Date dmin;
    Date dmax;
    Date dat;
    int sign;
    int i;

    if (dat1 < dat2)
    {
        sign = -1;
        dmin = dat1;
        dmax = dat2;
    }
    else
    {
        sign = +1;
        dmin = dat2;
        dmax = dat1;
    }

    for (i = 0, dat = dmin; dat < dmax; ++dat, ++i) {}

    return sign*i;    
}

Date& Date::operator++() // prefix
{
    return *this += 1;
}

Date Date::operator++(int) // postfix
{
    Date dtemp = *this;
    *this += 1;
    return dtemp;
}

Date& Date::operator--() // prefix
{
    return *this -= 1;
}

Date Date::operator--(int) // postfix
{
    Date dtemp = *this;
    *this -= 1;
    return dtemp;
}

Date& Date::operator+=(int n)
{
    return *this = *this + n;
}

Date& Date::operator-=(int n)
{
    return *this = *this - n;
}

//-----------------------------------------------------------------------
// Relational operators
//-----------------------------------------------------------------------

bool operator==(Date dat1, Date dat2)
{
    return dat1._year == dat2._year
        && dat1._month == dat2._month
        && dat1._day == dat2._day;
}

bool operator!=(Date dat1, Date dat2)
{
    return !(dat1 == dat2);
}

bool operator<(Date dat1, Date dat2)
{
    if (dat1._year < dat2._year) return true;
    else if (dat1._year == dat2._year)
    {
        if (dat1._month < dat2._month) return true;
        else if  (dat1._month == dat2._month) return dat1._day < dat2._day;
        else return false;
    }
    else return false;
}

bool operator<=(Date dat1, Date dat2)
{
    return dat1 < dat2 || dat1 == dat2;
}

bool operator>(Date dat1, Date dat2)
{
    return !(dat1 <= dat2);
}

bool operator>=(Date dat1, Date dat2)
{
    return !(dat1 < dat2);
}

//-----------------------------------------------------------------------
// IO operators
//-----------------------------------------------------------------------

// Display under readable form
ostream& operator<<(ostream& os, Date dat)
{
    os << Month[dat._month] << ' ' << dat._day << ", " << dat._year;
    return os;
}

// Read a date as three integers separated by /
istream& operator>>(istream& is, Date& dat)
{
    int y;
    int m;
    int d;
    char sep1, sep2;

    // If we cannot read three integers in a row, separated by /
    // fail with an exception */
    if (is >> y >> sep1 >> m >> sep2 >> d) 
    {
	if (sep1 != '/' || sep2 != '/') throw Date::Bad_Input();
	dat = Date(y, m, d);
    }
    else throw Date::Bad_Input();
    return is;
}

